module fact

use partition
use context
implicit none
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  factorization module: loc_fact + gb_fact  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


    !! 1. local and global UL structure: if NO compression, only ary11 is allocated
    type :: UL_struct
        integer                     :: rk1 = 0, rk2 = 0
        complex(kind=4),allocatable :: ary11(:,:), ary12(:,:), ary21(:,:), ary22(:,:)
        complex(kind=4),allocatable ::    Q1(:,:),    R1(:,:),    Q2(:,:),    R2(:,:)
    end type UL_struct
    type(UL_struct),allocatable :: locUL(:), gbUL(:)


    !! 2. local and global UR and LL structure: if NO compression, only ary is allocated
    type :: UR_struct
        integer                     :: rk = 0
        complex(kind=4),allocatable :: ary(:,:), Q1(:,:), Q2(:,:), R(:,:)
    end type UR_struct
    type(UR_struct),allocatable :: locUR(:), locLL(:), gbUR(:), gbLL(:)    


    !! 3. pivoting structure: if NO compression, only ary1 is allocated
    type :: IPIV_struct
        integer,allocatable :: ary1(:), ary2(:)
    end type IPIV_struct
    type(IPIV_struct),allocatable :: locIPIV(:), gbIPIV(:)


    !! 4. define factor for gbLR: needs to be redistributed over another process grid
    type :: redist_struct
        complex(kind=4),allocatable :: ary1(:,:), ary2(:,:)    ! ary1 on grid1, ary2 on grid2
    end type redist_struct


    !! 5. BLACS parameters
    integer,parameter :: RSRC = 0, CSRC = 0, MB = 64, NB = 64


contains


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!  The wrapping subroutine for factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine factorization(A,slvl,Lpar,rkmin,tol,mr2d,flag,ctxt,nprow,npcol,myrow,mycol,mynodes, & 
                             lenlocTT,lengbTT,locTT,gbTT,locTlabel,gbTlabel,locNB,gbNB,locUL,locUR,locLL,locIPIV,gbUL,gbUR,gbLL,gbIPIV)

        type(CSR)                       :: A
        real(kind=4)                    :: tol
        integer                         :: slvl, Lpar, rkmin, mr2d, flag, lenlocTT, lengbTT
        integer                         :: ctxt(Lpar), nprow(Lpar), npcol(Lpar), myrow(Lpar), mycol(Lpar), mynodes(Lpar)
        integer                         :: locTlabel(lenlocTT+1), gbTlabel(lengbTT+1)
        type(TT_struct)                 :: locTT(lenlocTT), gbTT(lengbTT)
        type(NB_struct)                 :: locNB(lenlocTT), gbNB(lengbTT)
        type(UL_struct),allocatable     :: locUL(:), gbUL(:)
        type(UR_struct),allocatable     :: locUR(:), locLL(:), gbUR(:), gbLL(:)
        type(IPIV_struct),allocatable   :: locIPIV(:), gbIPIV(:)

        ! dynamically allocated structure
        type(redist_struct),allocatable :: gbLR(:)
        integer         :: stat = 0
        real(kind=8)    :: flops = 0.0D+0, storage = 0.0D+0

        !! 1. local factorization
        call loc_fact( A, slvl, Lpar, rkmin, tol, flag, lengbTT, lenlocTT, locTT, locTlabel, locNB, locUL, locUR, locLL, locIPIV, gbLR, stat, flops, storage )
        deallocate( A%ind, A%col, A%val )
        if (flag == 1) then
            write(*,*) "local factorization is done!" 
        end if


        !! 2. global factorization
        call gb_fact( slvl, rkmin, tol, mr2d, flag, lengbTT, gbTT, gbTlabel, gbNB, ctxt, nprow, npcol, myrow, mycol, mynodes, gbUL, gbUR, gbLL, gbLR, gbIPIV, stat, flops, storage )
        deallocate( gbLR )
        if (flag == 1) then
            write(*,*) "global factorization is done!"
        end if


        !! 3. statistics if needed
        if (stat /= 0) then
            call dgsum2d( ctxt(1), 'A', 'I', 1, 1, flops,   1, 0, 0 )
            call dgsum2d( ctxt(1), 'A', 'I', 1, 1, storage, 1, 0, 0 )   
            if (flag == 1) then
                write(*,*) "total factorization flops is: ", flops
                write(*,*) "total storage is: ", storage
            end if             
        end if


        return
    end subroutine factorization



    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  1. deallocating subroutines for all factors  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine factors_dealloc( lenlocTT, locUL, locUR, locLL, locIPIV, Lvl, gbUL, gbUR, gbLL, gbIPIV )

        integer           :: lenlocTT, Lvl
        type(UL_struct)   :: locUL(lenlocTT), gbUL(Lvl)
        type(UR_struct)   :: locUR(lenlocTT), locLL(lenlocTT), gbUR(Lvl), gbLL(Lvl)
        type(IPIV_struct) :: locIPIV(lenlocTT), gbIPIV(Lvl)

        integer :: Ntmp, level

        do Ntmp = 1,lenlocTT
            ! locUL(Ntmp)
            if (allocated(locUL(Ntmp)%ary11)) then
                deallocate(locUL(Ntmp)%ary11)
            end if
            if (allocated(locUL(Ntmp)%ary22)) then
                deallocate(locUL(Ntmp)%ary22)
            end if
            if (allocated(locUL(Ntmp)%Q1)) then
                deallocate(locUL(Ntmp)%Q1)
            end if
            if (allocated(locUL(Ntmp)%Q2)) then
                deallocate(locUL(Ntmp)%Q2)
            end if
            if (allocated(locUL(Ntmp)%R1)) then
                deallocate(locUL(Ntmp)%R1)
            end if
            if (allocated(locUL(Ntmp)%R2)) then
                deallocate(locUL(Ntmp)%R2)
            end if
            ! locUR(Ntmp)
            if (allocated(locUR(Ntmp)%ary)) then
                deallocate(locUR(Ntmp)%ary)
            end if
            if (allocated(locUR(Ntmp)%Q1)) then
                deallocate(locUR(Ntmp)%Q1)
            end if
            if (allocated(locUR(Ntmp)%Q2)) then
                deallocate(locUR(Ntmp)%Q2)
            end if
            if (allocated(locUR(Ntmp)%R)) then
                deallocate(locUR(Ntmp)%R)
            end if
            ! locLL(Ntmp)
            if (allocated(locLL(Ntmp)%ary)) then
                deallocate(locLL(Ntmp)%ary)
            end if
            if (allocated(locLL(Ntmp)%Q1)) then
                deallocate(locLL(Ntmp)%Q1)
            end if
            if (allocated(locLL(Ntmp)%Q2)) then
                deallocate(locLL(Ntmp)%Q2)
            end if
            if (allocated(locLL(Ntmp)%R)) then
                deallocate(locLL(Ntmp)%R)
            end if
            ! locIPIV(Ntmp)
            if (allocated(locIPIV(Ntmp)%ary1)) then
                deallocate(locIPIV(Ntmp)%ary1)
            end if
            if (allocated(locIPIV(Ntmp)%ary2)) then
                deallocate(locIPIV(Ntmp)%ary2)
            end if
        end do

        do level = 1,Lvl
            ! gbUL(level)
            if (allocated(gbUL(level)%ary11)) then
                deallocate(gbUL(level)%ary11)
            end if
            if (allocated(gbUL(level)%ary22)) then
                deallocate(gbUL(level)%ary22)
            end if
            if (allocated(gbUL(level)%Q1)) then
                deallocate(gbUL(level)%Q1)
            end if
            if (allocated(gbUL(level)%Q2)) then
                deallocate(gbUL(level)%Q2)
            end if
            if (allocated(gbUL(level)%R1)) then
                deallocate(gbUL(level)%R1)
            end if
            if (allocated(gbUL(level)%R2)) then
                deallocate(gbUL(level)%R2)
            end if
            ! gbUR(level)
            if (allocated(gbUR(level)%ary)) then
                deallocate(gbUR(level)%ary)
            end if
            if (allocated(gbUR(level)%Q1)) then
                deallocate(gbUR(level)%Q1)
            end if
            if (allocated(gbUR(level)%Q2)) then
                deallocate(gbUR(level)%Q2)
            end if
            if (allocated(gbUR(level)%R)) then
                deallocate(gbUR(level)%R)
            end if
            ! gbLL(level)
            if (allocated(gbLL(level)%ary)) then
                deallocate(gbLL(level)%ary)
            end if
            if (allocated(gbLL(level)%Q1)) then
                deallocate(gbLL(level)%Q1)
            end if
            if (allocated(gbLL(level)%Q2)) then
                deallocate(gbLL(level)%Q2)
            end if
            if (allocated(gbLL(level)%R)) then
                deallocate(gbLL(level)%R)
            end if
            ! gbIPIV(level)
            if (allocated(gbIPIV(level)%ary1)) then
                deallocate(gbIPIV(level)%ary1)
            end if
            if (allocated(gbIPIV(level)%ary2)) then
                deallocate(gbIPIV(level)%ary2)
            end if
        end do

        return
    end subroutine factors_dealloc

    

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  2. local factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine loc_fact( A, slvl, Lpar, rkmin, tol, flag, lengbTT, lenlocTT, locTT, locTlabel, locNB, locUL, locUR, locLL, locIPIV, gbLR, stat, flops, storage )

        type(CSR)                          :: A
        integer                            :: slvl, Lpar, rkmin, flag, lengbTT, lenlocTT
        real(kind=4)                       :: tol
        type(TT_struct)                    :: locTT(lenlocTT)
        integer                            :: locTlabel(lenlocTT+1)
        type(NB_struct),target             :: locNB(lenlocTT)
        type(UL_struct),allocatable        :: locUL(:)
        type(UR_struct),allocatable        :: locUR(:), locLL(:)
        type(IPIV_struct),allocatable      :: locIPIV(:)
        type(redist_struct),allocatable    :: gbLR(:)
        integer                            :: stat
        real(kind=8)                       :: flops, storage

        integer                 :: level, Ntmp, Nhead, Ntail, Ptmp, Phead, Ptail, sz1, sz2, szUL, szLR, sz, ii, jj, counter, info
        type(NB_struct),pointer :: tail
        type :: Schur_struct
            complex(kind=4),allocatable :: ary(:,:)
        end type Schur_struct
        type(Schur_struct),allocatable :: Schur(:)

        allocate( locUL(lenlocTT), locUR(lenlocTT), locLL(lenlocTT), Schur(lenlocTT), locIPIV(lenlocTT), gbLR(lengbTT) )

        do Ntmp = 1,lenlocTT

            level = locTT(Ntmp)%LL
            Nhead = locTlabel(Ntmp)
            Ntail = locTlabel(Ntmp+1)-1
            szUL  = Ntail - Nhead + 1
            if (level <= slvl-Lpar+1) then
                sz1 = szUL / 2
                sz2 = szUL - sz1
            end if
            if (locNB(Ntmp)%Ohead /= 0) then
                szLR = 0
                tail => locNB(Ntmp)
                do while (associated(tail))
                    szLR = szLR + tail%Otail - tail%Ohead + 1
                    tail => tail%next
                end do
                sz = szLR
            end if

            ! allocating memory for leaf nodes factors: locUL, locUR, locLL and Schur complement
            if (locTT(Ntmp)%kids(1) == 0) then
                if (level <= slvl-Lpar+1) then
                    allocate( locUL(Ntmp)%ary11(sz1,sz1), locUL(Ntmp)%ary12(sz1,sz2), locUL(Ntmp)%ary21(sz1,sz2), locUL(Ntmp)%ary22(sz2,sz2) )
                    allocate( locIPIV(Ntmp)%ary1(sz1), locIPIV(Ntmp)%ary2(sz2) )
                    locUL(Ntmp)%ary11 = ZERO
                    locUL(Ntmp)%ary12 = ZERO
                    locUL(Ntmp)%ary21 = ZERO
                    locUL(Ntmp)%ary22 = ZERO
                else
                    allocate(locUL(Ntmp)%ary11(szUL,szUL),locIPIV(Ntmp)%ary1(szUL))
                    locUL(Ntmp)%ary11  = ZERO
                end if
                if (locNB(Ntmp)%Ohead /= 0) then                    
                    allocate( locUR(Ntmp)%ary(szUL,szLR), locLL(Ntmp)%ary(szUL,szLR) )
                    locUR(Ntmp)%ary = ZERO
                    locLL(Ntmp)%ary = ZERO
                    if (Ntmp /= lenlocTT) then
                        allocate( Schur(Ntmp)%ary(szLR,szLR) )
                        Schur(Ntmp)%ary = ZERO
                    else
                        allocate( gbLR(flag)%ary1(szLR,szLR) )
                        gbLR(flag)%ary1 = ZERO
                    end if
                end if
            end if
           
            !!!!!!!!!!!!!!!!!!!!!!!!  adding A  !!!!!!!!!!!!!!!!!!!!!!!!!!!!
            if (level <= slvl-Lpar+1) then 
                ! ary11
                do ii = Nhead,Nhead+sz1-1
                    do jj = A%ind(ii),A%ind(ii+1)-1
                        if (A%col(jj) >= Nhead .and. A%col(jj) < Nhead+sz1) then
                            locUL(Ntmp)%ary11( ii-Nhead+1, A%col(jj)-Nhead+1 ) = locUL(Ntmp)%ary11( ii-Nhead+1, A%col(jj)-Nhead+1 ) + A%val(jj)
                        end if
                    end do
                end do   
                ! ary12      
                do ii = Nhead,Nhead+sz1-1
                    do jj = A%ind(ii),A%ind(ii+1)-1
                        if (A%col(jj) >= Nhead+sz1 .and. A%col(jj) <= Ntail) then
                            locUL(Ntmp)%ary12( ii-Nhead+1, A%col(jj)-Nhead-sz1+1 ) = locUL(Ntmp)%ary12( ii-Nhead+1, A%col(jj)-Nhead-sz1+1 ) + A%val(jj)
                        end if
                    end do
                end do
                ! ary21
                do ii = Nhead+sz1,Ntail
                    do jj = A%ind(ii),A%ind(ii+1)-1
                        if (A%col(jj) >= Nhead .and. A%col(jj) < Nhead+sz1) then
                            locUL(Ntmp)%ary21( A%col(jj)-Nhead+1, ii-Nhead-sz1+1 ) = locUL(Ntmp)%ary21( A%col(jj)-Nhead+1, ii-Nhead-sz1+1 ) + A%val(jj)
                        end if
                    end do
                end do
                ! ary22
                do ii = Nhead+sz1,Ntail
                    do jj = A%ind(ii),A%ind(ii+1)-1
                        if (A%col(jj) >= Nhead+sz1 .and. A%col(jj) <= Ntail) then
                            locUL(Ntmp)%ary22( ii-Nhead-sz1+1, A%col(jj)-Nhead-sz1+1 ) = locUL(Ntmp)%ary22( ii-Nhead-sz1+1, A%col(jj)-Nhead-sz1+1 ) + A%val(jj)
                        end if
                    end do
                end do
            else
                ! ary11
                do ii = Nhead,Ntail
                    do jj = A%ind(ii),A%ind(ii+1)-1
                        if (A%col(jj) >= Nhead .and. A%col(jj) <= Ntail) then
                            locUL(Ntmp)%ary11( ii-Nhead+1, A%col(jj)-Nhead+1 ) = locUL(Ntmp)%ary11( ii-Nhead+1, A%col(jj)-Nhead+1 ) + A%val(jj)
                        end if
                    end do
                end do
            end if
            if (stat /= 0) then
                flops = flops + dble(szUL)**2.0
            end if

            ! if Ntmp has neighbors
            if (locNB(Ntmp)%Ohead /= 0) then  
                counter = 1
                tail => locNB(Ntmp)
                do while (associated(tail))
                    Phead = tail%Ohead
                    Ptail = tail%Otail
                    ! locUR copy
                    do ii = Nhead,Ntail
                        do jj = A%ind(ii),A%ind(ii+1)-1
                            if (A%col(jj) >= Phead .and. A%col(jj) <= Ptail) then
                                locUR(Ntmp)%ary( ii-Nhead+1, A%col(jj)-Phead+counter ) = locUR(Ntmp)%ary( ii-Nhead+1, A%col(jj)-Phead+counter ) + A%val(jj)
                            end if
                        end do
                    end do
                    ! locLL copy: take the TRANSPOSE !!!!!
                    do ii = Phead,Ptail
                        do jj = A%ind(ii),A%ind(ii+1)-1
                            if (A%col(jj) >= Nhead .and. A%col(jj) <= Ntail) then
                                locLL(Ntmp)%ary( A%col(jj)-Nhead+1, ii-Phead+counter ) = locLL(Ntmp)%ary( A%col(jj)-Nhead+1, ii-Phead+counter ) + A%val(jj)
                            end if
                        end do
                    end do
                    counter = counter + Ptail-Phead+1
                    tail => tail%next
                end do
                if (stat /= 0) then
                    flops = flops + dble(szUL) * dble(szLR) * 2.0
                end if
                !! reaching the root node of the local tree. copy A to gbLR(flag), which contains all interface matrix
                if (Ntmp == lenlocTT) then
                    do ii = Ntail+1,Ntail+szLR
                        do jj = A%ind(ii), A%ind(ii+1)-1
                            if (A%col(jj) > Ntail .and. A%col(jj) <= Ntail+szLR) then
                                gbLR(flag)%ary1( ii-Ntail, A%col(jj)-Ntail ) = gbLR(flag)%ary1( ii-Ntail, A%col(jj)-Ntail ) + A%val(jj)
                            end if
                        end do
                    end do
                    if (stat /= 0) then
                        flops = flops + dble(szLR)**2.0
                    end if
                end if
            end if

            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            ! locUL factorization
            if (level <= slvl-Lpar+1) then

                ! compression of ary12 and ary21
                call loc_compr( sz1, sz2, locUL(Ntmp)%ary12, rkmin, tol, locUL(Ntmp)%Q1, locUL(Ntmp)%R1, locUL(Ntmp)%rk1 )
                deallocate(locUL(Ntmp)%ary12)
                ! write(*,*) "local compr_rate: flag = ", flag, ", level = ", level, ", M = ", sz1, ", N = ", sz2, ", rank = ", locUL(Ntmp)%rk1
                if (stat /= 0) then
                    flops = flops + 4.0*dble(sz1)*dble(sz2)*dble(locUL(Ntmp)%rk1) - 2.0*dble(locUL(Ntmp)%rk1)**2.0*dble(szUL) + 4.0*dble(locUL(Ntmp)%rk1)**3.0/3.0
                    storage = storage + dble(szUL) * dble(locUL(Ntmp)%rk1)
                end if

                call loc_compr( sz1, sz2, locUL(Ntmp)%ary21, rkmin, tol, locUL(Ntmp)%Q2, locUL(Ntmp)%R2, locUL(Ntmp)%rk2 )
                deallocate(locUL(Ntmp)%ary21)
                ! write(*,*) "local compr_rate: flag = ", flag, ", level = ", level, ", M = ", sz1, ", N = ", sz2, ", rank = ", locUL(Ntmp)%rk2
                if (stat /= 0) then
                    flops = flops + 4.0*dble(sz1)*dble(sz2)*dble(locUL(Ntmp)%rk2) - 2.0*dble(locUL(Ntmp)%rk2)**2.0*dble(szUL) + 4.0*dble(locUL(Ntmp)%rk2)**3.0/3.0
                    storage = storage + dble(szUL) * dble(locUL(Ntmp)%rk2)
                end if

                ! ary11 factorization
                call cgetrf( sz1, sz1, locUL(Ntmp)%ary11, sz1, locIPIV(Ntmp)%ary1, info )
                if (stat /= 0) then
                    flops = flops + 2.0 * dble(sz1)**3.0 / 3.0 - dble(sz1)**2.0 / 2.0 - dble(sz1) / 6.0
                    storage = storage + dble(sz1)**2.0 + dble(sz1)
                end if                

                ! updating Q1 and Q2
                call claswp( locUL(Ntmp)%rk1, locUL(Ntmp)%Q1, sz1, 1, sz1, locIPIV(Ntmp)%ary1, 1 )
                call ctrtrs( 'L', 'N', 'U', sz1, locUL(Ntmp)%rk1, locUL(Ntmp)%ary11, sz1, locUL(Ntmp)%Q1, sz1, info )
                call ctrtrs( 'U', 'T', 'N', sz1, locUL(Ntmp)%rk2, locUL(Ntmp)%ary11, sz1, locUL(Ntmp)%Q2, sz1, info )
                if (stat /= 0) then
                    flops = flops + dble(sz1)**2.0 * ( dble(locUL(Ntmp)%rk1) + dble(locUL(Ntmp)%rk2) )
                end if

                ! BLAS 3 updating ary22
                call locmm_compr( locUL(Ntmp)%Q1, sz1, locUL(Ntmp)%R1, locUL(Ntmp)%rk1, locUL(Ntmp)%Q2, sz1, locUL(Ntmp)%R2, locUL(Ntmp)%rk2, locUL(Ntmp)%ary22, sz2 )
                if (stat /= 0) then
                    flops = flops + dble(locUL(Ntmp)%rk1) * dble(locUL(Ntmp)%rk2) * 2.0*dble(sz1) + dble(locUL(Ntmp)%rk2) * dble(sz2) * 2.0*dble(locUL(Ntmp)%rk1) &
                                  + dble(sz2) * dble(sz2) * 2.0*dble(locUL(Ntmp)%rk2)
                end if

                ! ary22 factorization
                call cgetrf( sz2, sz2, locUL(Ntmp)%ary22, sz2, locIPIV(Ntmp)%ary2, info )
                if (stat /= 0) then
                    flops = flops + 2.0 * dble(sz2)**3.0 / 3.0 - dble(sz2)**2.0 / 2.0 - dble(sz2) / 6.0
                    storage = storage + dble(sz2)**2.0 + dble(sz2)
                end if                

            else
                
                ! factorization of the entire ary11                
                call cgetrf( szUL, szUL, locUL(Ntmp)%ary11, szUL, locIPIV(Ntmp)%ary1, info )
                if (stat /= 0) then
                    flops = flops + 2.0 * dble(szUL)**3.0 / 3.0 - dble(szUL)**2.0 / 2.0 - dble(szUL) / 6.0
                    storage = storage + dble(szUL)**2.0 + dble(szUL)
                end if            

            end if

            ! if the current node has neighbors
            if (locNB(Ntmp)%Ohead /= 0) then              

                if (level <= slvl-Lpar+1) then

                    ! compression of UR and LL
                    call loc_compr( szUL, szLR, locUR(Ntmp)%ary, rkmin, tol, locUR(Ntmp)%Q1, locUR(Ntmp)%R, locUR(Ntmp)%rk )
                    deallocate(locUR(Ntmp)%ary)
                    ! write(*,*) "local compr_rate: flag = ", flag, ", level = ", level, ", M = ", szUL, ", N = ", szLR, ", rank = ", locUR(Ntmp)%rk
                    if (stat /= 0) then
                        flops = flops + 4.0*dble(szUL)*dble(szLR)*dble(locUR(Ntmp)%rk) - 2.0*dble(locUR(Ntmp)%rk)**2.0*(dble(szUL)+dble(szLR)) & 
                                      + 4.0*dble(locUR(Ntmp)%rk)**3.0/3.0
                        storage = storage + dble(locUR(Ntmp)%rk) * ( dble(szUL) + dble(szLR) )
                    end if

                    call loc_compr( szUL, szLR, locLL(Ntmp)%ary, rkmin, tol, locLL(Ntmp)%Q1, locLL(Ntmp)%R, locLL(Ntmp)%rk )
                    deallocate(locLL(Ntmp)%ary)
                    ! write(*,*) "local compr_rate: flag = ", flag, ", level = ", level, ", M = ", szUL, ", N = ", szLR, ", rank = ", locLL(Ntmp)%rk
                    if (stat /= 0) then
                        flops = flops + 4.0*dble(szUL)*dble(szLR)*dble(locLL(Ntmp)%rk) - 2.0*dble(locLL(Ntmp)%rk)**2.0*(dble(szUL) + dble(szLR)) & 
                                      + 4.0*dble(locLL(Ntmp)%rk)**3.0/3.0
                        storage = storage + dble(locLL(Ntmp)%rk) * ( dble(szUL) + dble(szLR) )
                    end if

                    ! two triangular solvers
                    call loctrs_compr( -1, 0, locUR(Ntmp)%rk, locUL(Ntmp), locIPIV(Ntmp), locUR(Ntmp)%Q1, szUL )
                    call loctrs_compr(  1, 1, locLL(Ntmp)%rk, locUL(Ntmp), locIPIV(Ntmp), locLL(Ntmp)%Q1, szUL )
                    if (stat /= 0) then
                        flops = flops + dble(szUL)**2.0 * ( dble(locUR(Ntmp)%rk) + dble(locLL(Ntmp)%rk) )
                    end if

                    ! BLAS 3 updating
                    if (Ntmp /= lenlocTT) then
                        call locmm_compr( locUR(Ntmp)%Q1, szUL, locUR(Ntmp)%R, locUR(Ntmp)%rk, locLL(Ntmp)%Q1, szUL, locLL(Ntmp)%R, locLL(Ntmp)%rk, Schur(Ntmp)%ary, szLR )
                    else
                        call locmm_compr( locUR(Ntmp)%Q1, szUL, locUR(Ntmp)%R, locUR(Ntmp)%rk, locLL(Ntmp)%Q1, szUL, locLL(Ntmp)%R, locLL(Ntmp)%rk, gbLR(flag)%ary1, szLR )
                    end if
                    if (stat /= 0) then
                        flops = flops + dble(locUR(Ntmp)%rk) * dble(locLL(Ntmp)%rk) * 2.0*dble(szUL) + dble(locLL(Ntmp)%rk) * dble(szLR) * 2.0*dble(locUR(Ntmp)%rk) &
                                      + dble(szLR) * dble(szLR) * 2.0*dble(locLL(Ntmp)%rk)
                    end if

                else

                    ! triangular solver for locUR and locLL
                    call claswp( szLR, locUR(Ntmp)%ary, szUL, 1, szUL, locIPIV(Ntmp)%ary1, 1 )
                    call ctrtrs( 'L', 'N', 'U', szUL, szLR, locUL(Ntmp)%ary11, szUL, locUR(Ntmp)%ary, szUL, info )
                    call ctrtrs( 'U', 'T', 'N', szUL, szLR, locUL(Ntmp)%ary11, szUL, locLL(Ntmp)%ary, szUL, info )
                    if (stat /= 0) then
                        flops = flops + dble(szUL)**2.0 * dble(szLR) * 2.0
                        storage = storage + dble(szUL) * dble(szLR) * 2.0
                    end if

                    ! BLAS 3: Schur complement
                    if (Ntmp /= lenlocTT) then
                        call cgemm( 'T', 'N', szLR, szLR, szUL, -ONE, locLL(Ntmp)%ary, szUL, locUR(Ntmp)%ary, szUL, ONE, Schur(Ntmp)%ary, szLR )
                    else 
                        call cgemm( 'T', 'N', szLR, szLR, szUL, -ONE, locLL(Ntmp)%ary, szUL, locUR(Ntmp)%ary, szUL, ONE, gbLR(flag)%ary1, szLR )
                    end if
                    if (stat /= 0) then
                        flops = flops + dble(szLR) * dble(szLR) * 2.0*dble(szUL)
                    end if

                end if
 
                !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  extend-add  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                if (Ntmp /= lenlocTT) then

                    Ptmp  = locTT(Ntmp)%parent
                    level = locTT(Ptmp)%LL
                    szUL  = locTlabel(Ptmp+1) - locTlabel(Ptmp)
                    if (level <= slvl-Lpar+1) then
                        sz1 = szUL / 2
                        sz2 = szUL - sz1
                    end if
                    if (locNB(Ptmp)%Ohead /= 0) then
                        szLR = 0
                        tail => locNB(Ptmp)
                        do while (associated(tail))
                            szLR = szLR + tail%Otail - tail%Ohead + 1
                            tail => tail%next
                        end do
                    end if

                    ! left kid to allocate its parent's memory
                    if (Ntmp == locTT(Ptmp)%kids(1)) then
                        if (level <= slvl-Lpar+1) then
                            allocate( locUL(Ptmp)%ary11(sz1,sz1), locUL(Ptmp)%ary12(sz1,sz2), locUL(Ptmp)%ary21(sz1,sz2), locUL(Ptmp)%ary22(sz2,sz2) )
                            allocate( locIPIV(Ptmp)%ary1(sz1), locIPIV(Ptmp)%ary2(sz2) )
                            locUL(Ptmp)%ary11 = ZERO
                            locUL(Ptmp)%ary12 = ZERO
                            locUL(Ptmp)%ary21 = ZERO
                            locUL(Ptmp)%ary22 = ZERO
                        else
                            allocate(locUL(Ptmp)%ary11(szUL,szUL),locIPIV(Ptmp)%ary1(szUL))
                            locUL(Ptmp)%ary11  = ZERO
                        end if
                        if (locNB(Ptmp)%Ohead /= 0) then                    
                            allocate( locUR(Ptmp)%ary(szUL,szLR), locLL(Ptmp)%ary(szUL,szLR) )
                            locUR(Ptmp)%ary = ZERO
                            locLL(Ptmp)%ary = ZERO
                            if (Ptmp /= lenlocTT) then
                                allocate( Schur(Ptmp)%ary(szLR,szLR) )
                                Schur(Ptmp)%ary = ZERO
                            else
                                allocate( gbLR(flag)%ary1(szLR,szLR) )
                                gbLR(flag)%ary1 = ZERO
                            end if
                        end if
                    end if

                    ! extend-add
                    if (Ptmp /= lenlocTT) then
                        call loc_extadd( level, slvl, Lpar, locUL(Ptmp), szUL, locUR(Ptmp)%ary, locLL(Ptmp)%ary, Schur(Ptmp)%ary, szLR, Schur(Ntmp)%ary, sz, Ntmp, Ptmp, locNB )
                    else
                        call loc_extadd( level, slvl, Lpar, locUL(Ptmp), szUL, locUR(Ptmp)%ary, locLL(Ptmp)%ary, gbLR(flag)%ary1, szLR, Schur(Ntmp)%ary, sz, Ntmp, Ptmp, locNB )
                    end if
                    deallocate(Schur(Ntmp)%ary)
                    if (stat /= 0) then
                        flops = flops + dble(sz)**2.0
                    end if

                end if

            end if
            
        end do
        deallocate(Schur)

        return
    end subroutine loc_fact 


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  local extend-add  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine loc_extadd( level, slvl, Lpar, UL, szUL, UR, LL, LR, szLR, U, szU, Ntmp, Ptmp, nbr )

        integer                :: level, slvl, Lpar, szUL, szLR, szU, Ntmp, Ptmp
        type(UL_struct)        :: UL
        complex(kind=4)        :: UR(szUL,szLR), LL(szUL,szLR), LR(szLR,szLR), U(szU,szU)
        type(NB_struct),target :: nbr(*)

        integer                  :: sz1, sz2, length, counter, ii, jj, xx, yy
        integer,allocatable      :: ind(:), map(:)
        type(NB_struct),pointer  :: tailN, tailP

        ! length of the ind and map
        length = 0
        tailN => nbr(Ntmp)%next
        do while (associated(tailN))
            length = length + 1
            tailN => tailN%next
        end do

        !! 1. UL
        if (level <= slvl-Lpar+1) then
            sz1 = szUL / 2
            sz2 = szUL - sz1
            UL%ary11 = UL%ary11 + U(1:sz1,1:sz1)
            UL%ary12 = UL%ary12 + U(1:sz1,sz1+1:szUL)
            do yy = 1,sz2
                do xx = 1,sz1
                    UL%ary21(xx,yy) = UL%ary21(xx,yy) + U(sz1+yy,xx)
                end do
            end do
            UL%ary22 = UL%ary22 + U(sz1+1:szUL,sz1+1:szUL)
        else
            UL%ary11 = UL%ary11 + U(1:szUL,1:szUL)
        end if

        if (length > 0) then

            ! ind: index of each block in U
            allocate(ind(length+1))
            ind(1) = szUL + 1
            tailN => nbr(Ntmp)%next
            do ii = 1,length
                ind(ii+1) = ind(ii) + tailN%Otail - tailN%Ohead + 1
                tailN => tailN%next
            end do

            ! mapping to the current frontal matrix
            allocate(map(length))
            tailN => nbr(Ntmp)%next
            do ii = 1,length
                counter = 1
                tailP => nbr(Ptmp)
                do while (associated(tailP))
                    if (tailN%Ohead >= tailP%Ohead .and. tailN%Otail <= tailP%Otail) then
                        map(ii) = tailN%Ohead - tailP%Ohead + counter
                        exit
                    end if
                    counter = counter + tailP%Otail - tailP%Ohead + 1
                    tailP => tailP%next
                end do
                tailN => tailN%next
            end do 

            !! 2. UR and LL
            do ii = 1,length
                UR( 1:szUL, map(ii):map(ii)+ind(ii+1)-ind(ii)-1 ) = UR( 1:szUL, map(ii):map(ii)+ind(ii+1)-ind(ii)-1 ) + U( 1:szUL, ind(ii):ind(ii+1)-1 )
                do yy = map(ii),map(ii)+ind(ii+1)-ind(ii)-1
                    do xx = 1,szUL
                        LL(xx,yy) = LL(xx,yy) + U( yy-map(ii)+ind(ii), xx )
                    end do
                end do
            end do

            !! 3. LR
            do jj = 1,length
                do ii = 1,length
                    LR( map(ii):map(ii)+ind(ii+1)-ind(ii)-1, map(jj):map(jj)+ind(jj+1)-ind(jj)-1 ) = & 
                    LR( map(ii):map(ii)+ind(ii+1)-ind(ii)-1, map(jj):map(jj)+ind(jj+1)-ind(jj)-1 ) + U( ind(ii):ind(ii+1)-1, ind(jj):ind(jj+1)-1 )
                end do
            end do

            deallocate(ind,map)

        end if

        return
    end subroutine loc_extadd  


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  local compression  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine loc_compr( M, N, A, rkmin, tol, Q, R, rk )

        complex(kind=4)             :: A(M,*)
        complex(kind=4),allocatable :: Q(:,:), R(:,:)
        real(kind=4)                :: tol
        integer                     :: M, N, rkmin, rk

        integer                     :: info, ii, jj, K
        integer,allocatable         :: jpvt(:)
        real(kind=4),allocatable    :: rwork(:)
        complex(kind=4),allocatable :: tau(:), work(:)

        K = min(M,N)

        ! QR factorization with column pivoting
        allocate( jpvt(N), tau(K), work(N+1), rwork(2*N) )
        jpvt = 0
        call cgeqp3( M, N, A, M, jpvt, tau, work, N+1, rwork, info )
        deallocate(work,rwork)

        ! estimating rank and compression
        if (rkmin > K) then
            rk = K
        else
            if ( abs(A(rkmin,rkmin)) < tol * abs(A(1,1)) ) then
                rk = rkmin
            else
                do ii = rkmin,K
                    if ( abs(A(ii,ii)) < tol * abs(A(1,1)) ) then
                        exit
                    else
                        rk = ii
                    end if
                end do
            end if
        end if        
        allocate(Q(M,rk),R(rk,N))
        Q = ZERO
        R = ZERO

        ! copy R
        do jj = 1,rk
            R(1:jj,jpvt(jj)) = A(1:jj,jj)
        end do
        do jj = rk+1,N
            R(1:rk,jpvt(jj)) = A(1:rk,jj)
        end do
        deallocate(jpvt)

        ! generate Q
        allocate(work(rk))
        call cungqr( M, rk, rk, A(:,1:rk), M, tau(1:rk), work, rk, info )
        deallocate(work,tau)
        Q = A(:,1:rk) 

        return
    end subroutine loc_compr 


    !!!!!!!!!!!!!!!!!!!!!!!!!!!  local UR and LL multiplication with compression  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine locmm_compr( Q1, LDQ1, R1, LDR1, Q2, LDQ2, R2, LDR2, A, LDA )

        integer                     :: LDQ1, LDR1, LDQ2, LDR2, LDA
        complex(kind=4)             :: Q1(LDQ1,*), R1(LDR1,*), Q2(LDQ2,*), R2(LDR2,*), A(LDA,*)

        integer                     :: rk1, sz1, rk2, sz2
        complex(kind=4),allocatable :: tmp1(:,:), tmp2(:,:)

        sz1 = LDQ1
        sz2 = LDA
        rk1 = LDR1
        rk2 = LDR2

        allocate( tmp1(rk2,rk1), tmp2(rk2,sz2) )
        tmp1 = ZERO
        tmp2 = ZERO
        call cgemm( 'T', 'N', rk2, rk1, sz1,  ONE, Q2,   sz1, Q1,   sz1, ONE, tmp1, rk2 )
        call cgemm( 'N', 'N', rk2, sz2, rk1,  ONE, tmp1, rk2, R1,   rk1, ONE, tmp2, rk2 )
        deallocate(tmp1)
        call cgemm( 'T', 'N', sz2, sz2, rk2, -ONE, R2,   rk2, tmp2, rk2, ONE, A,    sz2 )
        deallocate(tmp2)

        return
    end subroutine locmm_compr 


    !!!!!!!!!!!!!!!!!!!!!!!!!!!  local solution of one RHS with compression  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine loctrs_compr( UPLO, TRANS, NRHS, UL, IPIV, B, LDB )

        integer                     :: UPLO, TRANS, NRHS
        type(UL_struct)             :: UL
        type(IPIV_struct)           :: IPIV
        integer                     :: LDB
        complex(kind=4)             :: B(LDB,*)

        integer                     :: sz1, sz2, rk1, rk2, info
        complex(kind=4),allocatable :: tmp(:,:)

        sz1 = size(UL%ary11,1)
        sz2 = size(UL%ary22,1)
        rk1 = UL%rk1
        rk2 = UL%rk2

        if (UPLO < 0) then        ! lower triangular system

            call claswp( NRHS, B(1:sz1,1:NRHS), sz1, 1, sz1, IPIV%ary1, 1 )
            call ctrtrs( 'L', 'N', 'U', sz1, NRHS, UL%ary11, sz1, B(1:sz1,1:NRHS), sz1, info )
            allocate(tmp(rk2,NRHS))
            tmp = ZERO
            call cgemm( 'T', 'N', rk2, NRHS, sz1,  ONE, UL%Q2, sz1, B(1:sz1,1:NRHS), sz1, ONE, tmp, rk2 )
            call cgemm( 'T', 'N', sz2, NRHS, rk2, -ONE, UL%R2, rk2, tmp, rk2, ONE, B(sz1+1:sz1+sz2,1:NRHS), sz2 )
            deallocate(tmp)
            call claswp( NRHS, B(sz1+1:sz1+sz2,1:NRHS), sz2, 1, sz2, IPIV%ary2, 1 )
            call ctrtrs( 'L', 'N', 'U', sz2, NRHS, UL%ary22, sz2, B(sz1+1:sz1+sz2,1:NRHS), sz2, info )

        elseif (UPLO > 0) then     ! upper triangular system

            if (TRANS == 0) then    ! NO transpose

                call ctrtrs( 'U', 'N', 'N', sz2, NRHS, UL%ary22, sz2, B(sz1+1:sz1+sz2,1:NRHS), sz2, info )
                allocate(tmp(rk1,NRHS))
                tmp = ZERO
                call cgemm( 'N', 'N', rk1, NRHS, sz2,  ONE, UL%R1, rk1, B(sz1+1:sz1+sz2,1:NRHS), sz2, ONE, tmp, rk1 )
                call cgemm( 'N', 'N', sz1, NRHS, rk1, -ONE, UL%Q1, sz1, tmp, rk1, ONE, B(1:sz1,1:NRHS), sz1 )
                deallocate(tmp)
                call ctrtrs( 'U', 'N', 'N', sz1, NRHS, UL%ary11, sz1, B(1:sz1,1:NRHS), sz1, info )

            else                    ! YES transpose

                call ctrtrs( 'U', 'T', 'N', sz1, NRHS, UL%ary11, sz1, B(1:sz1,1:NRHS), sz1, info )
                allocate(tmp(rk1,NRHS))
                tmp = ZERO
                call cgemm( 'T', 'N', rk1, NRHS, sz1,  ONE, UL%Q1, sz1, B(1:sz1,1:NRHS), sz1, ONE, tmp, rk1 )
                call cgemm( 'T', 'N', sz2, NRHS, rk1, -ONE, UL%R1, rk1, tmp, rk1, ONE, B(sz1+1:sz1+sz2,1:NRHS), sz2 )
                deallocate(tmp)
                call ctrtrs( 'U', 'T', 'N', sz2, NRHS, UL%ary22, sz2, B(sz1+1:sz1+sz2,1:NRHS), sz2, info )

            end if

        end if

        return
    end subroutine loctrs_compr 



    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  3. global factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine gb_fact( slvl, rkmin, tol, mr2d, flag, lengbTT, gbTT, gbTlabel, gbNB, ctxt, nprow, npcol, myrow, mycol, mynodes, gbUL, gbUR, gbLL, gbLR, gbIPIV, stat, flops, storage )

        integer                         :: slvl, rkmin, mr2d, flag, lengbTT
        real(kind=4)                    :: tol
        type(TT_struct)                 :: gbTT(lengbTT)
        integer                         :: gbTlabel(lengbTT+1)
        type(NB_struct),target          :: gbNB(lengbTT)
        integer                         :: ctxt(*), nprow(*), npcol(*), myrow(*), mycol(*), mynodes(*)
        type(redist_struct)             :: gbLR(lengbTT)
        type(UL_struct),allocatable     :: gbUL(:)
        type(UR_struct),allocatable     :: gbUR(:), gbLL(:)
        type(IPIV_struct),allocatable   :: gbIPIV(:)
        integer                         :: stat
        real(kind=8)                    :: flops, storage

        integer                 :: Lvl, level, Ntmp, kid, szUL, szLR, sz1, sz2, sz, LOCR, LOCC, ii, jj, kk, tmp, rsd, info
        integer,external        :: numroc
        integer                 :: descIPIV(9,2), descUL(9,6), descUR(9,3), descLL(9,3), descLR(9), desc1(9), desc2(9)
        integer                 :: IWORK(1)
        type(NB_struct),pointer :: tail


        Lvl = gbTT(1)%LL - 1    !! CAUTION: Lpar-1
        allocate( gbUL(Lvl), gbUR(Lvl), gbLL(Lvl), gbIPIV(Lvl) )

        do level = Lvl,1,-1

            Ntmp = mynodes(level)
          ! call blacs_barrier(ctxt(level),'a')
          
            if (flag == 1) then
                write(*,*) "level = ", level
            end if
 
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!  allocating memory for the factors, set up descriptors  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
            szUL = gbTlabel(Ntmp+1) - gbTlabel(Ntmp)

            if (level <= slvl) then

                sz1 = szUL / 2
                sz2 = szUL - sz1

                LOCR = numroc( sz1, MB, myrow(level), RSRC, nprow(level) )
                LOCC = numroc( sz1, NB, mycol(level), CSRC, npcol(level) )
                allocate( gbUL(level)%ary11(LOCR,LOCC), gbIPIV(level)%ary1(LOCR+MB) )
                gbUL(level)%ary11 = ZERO
                call descset( descUL(:,1), sz1, sz1, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                call descset( descIPIV(:,1), sz1+MB*nprow(level), 1, MB, 1, RSRC, mycol(level), ctxt(level), MB+LOCR )

                LOCR = numroc( sz1, MB, myrow(level), RSRC, nprow(level) )
                LOCC = numroc( sz2, NB, mycol(level), CSRC, npcol(level) )
                allocate( gbUL(level)%ary12(LOCR,LOCC), gbUL(level)%ary21(LOCR,LOCC) )
                gbUL(level)%ary12 = ZERO
                gbUL(level)%ary21 = ZERO
                call descset( descUL(:,2), sz1, sz2, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                descUL(:,4) = descUL(:,2)

                LOCR = numroc( sz2, MB, myrow(level), RSRC, nprow(level) )
                LOCC = numroc( sz2, NB, mycol(level), CSRC, npcol(level) )
                allocate( gbUL(level)%ary22(LOCR,LOCC), gbIPIV(level)%ary2(LOCR+MB) )
                gbUL(level)%ary22 = ZERO
                call descset( descUL(:,6), sz2, sz2, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                call descset( descIPIV(:,2), sz2+MB*nprow(level), 1, MB, 1, RSRC, mycol(level), ctxt(level), MB+LOCR )

            else

                LOCR = numroc( szUL, MB, myrow(level), RSRC, nprow(level) )
                LOCC = numroc( szUL, NB, mycol(level), CSRC, npcol(level) )
                allocate( gbUL(level)%ary11(LOCR,LOCC), gbIPIV(level)%ary1(LOCR+MB) )
                gbUL(level)%ary11  = ZERO
                call descset( descUL(:,1), szUL, szUL, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                call descset( descIPIV(:,1), szUL+MB*nprow(level), 1, MB, 1, RSRC, mycol(level), ctxt(level), MB+LOCR )

            end if

            if (Ntmp /= lengbTT) then
 
                szLR = 0
                tail => gbNB(Ntmp)
                do while (associated(tail))
                    szLR = szLR + tail%Otail - tail%Ohead + 1
                    tail => tail%next
                end do

                LOCR = numroc( szUL, MB, myrow(level), RSRC, nprow(level) )
                LOCC = numroc( szLR, NB, mycol(level), CSRC, npcol(level) )
                allocate( gbUR(level)%ary(LOCR,LOCC), gbLL(level)%ary(LOCR,LOCC) ) 
                gbUR(level)%ary = ZERO
                gbLL(level)%ary = ZERO
                call descset( descUR(:,1), szUL, szLR, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                descLL(:,1) = descUR(:,1)
                
                LOCR = numroc( szLR, MB, myrow(level), RSRC, nprow(level) )
                LOCC = numroc( szLR, NB, mycol(level), CSRC, npcol(level) )
                allocate(gbLR(Ntmp)%ary1(LOCR,LOCC))
                gbLR(Ntmp)%ary1 = ZERO
                call descset( descLR, szLR, szLR, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )

            end if


            !!!!!!!!!!!!!!!!!!!!!!!!!  redistribute the gbLR to the new context and then EXTEND-ADD  !!!!!!!!!!!!!!!!!!!!!!!!!!!!
            do kk = 1,2

                kid = gbTT(Ntmp)%kids(kk)

                !! size of the kid
                sz = 0
                tail => gbNB(kid)
                do while (associated(tail))
                    sz = sz + tail%Otail - tail%Ohead + 1
                    tail => tail%next
                end do

                !! desc1
                if (kk == 1) then    ! left kid
                    if (flag <= gbTT(Ntmp)%kids(1)) then
                        LOCR = numroc( sz, MB, myrow(level+1), RSRC, nprow(level+1) )
                        call descset( desc1, sz, sz, MB, NB, RSRC, CSRC, ctxt(level+1), max(1,LOCR) )
                    else
                        call descset( desc1, sz, sz, MB, NB, RSRC, CSRC, -1, 1 )
                    end if
                else    ! right kid
                    if (flag <= gbTT(Ntmp)%kids(1)) then
                        call descset( desc1, sz, sz, MB, NB, RSRC, CSRC, -1, 1 )
                    else
                        LOCR = numroc( sz, MB, myrow(level+1), RSRC, nprow(level+1) )
                        call descset( desc1, sz, sz, MB, NB, RSRC, CSRC, ctxt(level+1), max(1,LOCR) )
                    end if
                end if

                !! desc2
                LOCR = numroc( sz, MB, myrow(level), RSRC, nprow(level) )
                LOCC = numroc( sz, NB, mycol(level), CSRC, npcol(level) )
                call descset( desc2, sz, sz, MB, NB, RSRC, CSRC, ctxt(level), max(1,LOCR) )
                allocate(gbLR(kid)%ary2(LOCR,LOCC))
                gbLR(kid)%ary2 = ZERO
 
                !! redistribute the gbLR(kid)%ary1 to gbLR(kid)%ary2 and then deallocate gbLR(kid)%ary1
                !! CAUTION: bug in pcgemr2d, not applicable to large matrix inter-context copy
                !! mr2d: the parameter controling entire redistribution or redistribution block by block
                if (mr2d /= 0) then
                    if (mod(Lvl-level,2) == 0) then    ! column redistribution: rectangular process grid
                        tmp = sz / NB
                        rsd = mod(sz,NB)
                        do jj = 1,tmp
                            call pcgemr2d( sz,  NB, gbLR(kid)%ary1, 1, (jj-1)*NB+1, desc1, gbLR(kid)%ary2, 1, (jj-1)*NB+1, desc2, ctxt(level) )
                        end do
                        if (rsd /= 0) then
                            call pcgemr2d( sz, rsd, gbLR(kid)%ary1, 1,    tmp*NB+1, desc1, gbLR(kid)%ary2, 1,    tmp*NB+1, desc2, ctxt(level) )
                        end if
                    else                               ! row redistribution: square process grid
                        tmp = sz / MB
                        rsd = mod(sz,MB)
                        do ii = 1,tmp
                            call pcgemr2d(  MB, sz, gbLR(kid)%ary1, (ii-1)*MB+1, 1, desc1, gbLR(kid)%ary2, (ii-1)*MB+1, 1, desc2, ctxt(level) )
                        end do
                        if (rsd /= 0) then
                            call pcgemr2d( rsd, sz, gbLR(kid)%ary1,    tmp*MB+1, 1, desc1, gbLR(kid)%ary2,    tmp*MB+1, 1, desc2, ctxt(level) )
                        end if
                    end if
                else
                            call pcgemr2d(  sz, sz, gbLR(kid)%ary1,           1, 1, desc1, gbLR(kid)%ary2,           1, 1, desc2, ctxt(level) )
                end if

                if (allocated(gbLR(kid)%ary1)) then
                    deallocate(gbLR(kid)%ary1)     
                end if

                !! parallel extend-add from gbLR(kid)%ary2 to UL(Ntmp), UR(Ntmp), LL(Ntmp) and LR(Ntmp)
                call pextadd( level, slvl, gbUL(level), descUL, gbUR(level)%ary, gbLL(level)%ary, descUR(:,1), gbLR(Ntmp)%ary1, descLR, gbLR(kid)%ary2, desc2, Ntmp, kid, gbNB )
                if (allocated(gbLR(kid)%ary2)) then
                    deallocate(gbLR(kid)%ary2)
                end if

                if (stat /= 0) then
                    if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                        flops = flops + 2.0 * dble(sz)**2.0
                    end if
                end if

            end do


            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  gbUL factorization  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            
            if (level <= slvl) then

                ! compression ary12 and ary21
                desc1 = descUL(:,2)
                call gbUL_compr( gbUL(level)%ary12, desc1, ctxt(level), nprow(level), npcol(level), myrow(level), mycol(level), rkmin, tol, gbUL(level)%Q1, descUL(:,2), gbUL(level)%R1, descUL(:,3), gbUL(level)%rk1 )
                deallocate(gbUL(level)%ary12)
                if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                    ! write(*,*) "global compr_rate: level = ", level, ", M = ", desc1(3), ", N = ", desc1(4), ", rank = ", gbUL(level)%rk1
                end if
                if (stat /= 0) then
                    if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                        flops = flops + 4.0*dble(sz1)*dble(sz2)*dble(gbUL(level)%rk1) - 2.0*dble(gbUL(level)%rk1)**2.0*dble(szUL) & 
                                      + 4.0*dble(gbUL(level)%rk1)**3.0/3.0 
                        storage = storage + dble(szUL) * dble(gbUL(level)%rk1)
                    end if
                end if      

                desc2 = descUL(:,4)
                call gbUL_compr( gbUL(level)%ary21, desc2, ctxt(level), nprow(level), npcol(level), myrow(level), mycol(level), rkmin, tol, gbUL(level)%Q2, descUL(:,4), gbUL(level)%R2, descUL(:,5), gbUL(level)%rk2 )
                deallocate(gbUL(level)%ary21)
                if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                    ! write(*,*) "global compr_rate: level = ", level, ", M = ", desc2(3), ", N = ", desc2(4), ", rank = ", gbUL(level)%rk2
                end if
                if (stat /= 0) then
                    if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                        flops = flops + 4.0*dble(sz1)*dble(sz2)*dble(gbUL(level)%rk2) - 2.0*dble(gbUL(level)%rk2)**2.0*dble(szUL) & 
                                      + 4.0*dble(gbUL(level)%rk2)**3.0/3.0
                        storage = storage + dble(szUL) * dble(gbUL(level)%rk2)
                    end if
                end if

                ! factorization of ary11
                call pcgetrf( sz1, sz1, gbUL(level)%ary11, 1, 1, descUL(:,1), gbIPIV(level)%ary1, info )
                if (stat /= 0) then
                    if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                        flops = flops + 2.0 * dble(sz1)**3.0 / 3.0 - dble(sz1)**2.0 / 2.0 - dble(sz1) / 6.0
                        storage = storage + dble(sz1)**2.0 + dble(sz1)
                    end if
                end if

                ! updating Q1 and Q2
                call pclapiv( 'F', 'R', 'C', sz1, gbUL(level)%rk1, gbUL(level)%Q1, 1, 1, descUL(:,2), gbIPIV(level)%ary1, 1, 1, descIPIV(:,1), IWORK )
                call pctrtrs( 'L', 'N', 'U', sz1, gbUL(level)%rk1, gbUL(level)%ary11, 1, 1, descUL(:,1), gbUL(level)%Q1, 1, 1, descUL(:,2), info )
                call pctrtrs( 'U', 'T', 'N', sz1, gbUL(level)%rk2, gbUL(level)%ary11, 1, 1, descUL(:,1), gbUL(level)%Q2, 1, 1, descUL(:,4), info )
                if (stat /= 0) then
                    if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                        flops = flops + dble(sz1)**2.0 * ( dble(gbUL(level)%rk1) + dble(gbUL(level)%rk2) )
                    end if
                end if

                ! BLAS 3 
                call gbULmm_compr(ctxt(level),nprow(level),npcol(level),myrow(level),mycol(level),gbUL(level)%Q1,descUL(:,2),gbUL(level)%R1,descUL(:,3),gbUL(level)%Q2,descUL(:,4),gbUL(level)%R2,descUL(:,5),gbUL(level)%ary22,descUL(:,6))
                if (stat /= 0) then
                    if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                        flops = flops + dble(gbUL(level)%rk1) * dble(gbUL(level)%rk2) * 2.0*dble(sz1) + dble(gbUL(level)%rk2) * dble(sz2) * 2.0*dble(gbUL(level)%rk1) &
                                      + dble(sz2) * dble(sz2) * 2.0*dble(gbUL(level)%rk2)
                    end if
                end if

                ! factorization of ary22
                call pcgetrf( sz2, sz2, gbUL(level)%ary22, 1, 1, descUL(:,6), gbIPIV(level)%ary2, info )
                if (stat /= 0) then
                    if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                        flops = flops + 2.0 * dble(sz2)**3.0 / 3.0 - dble(sz2)**2.0 / 2.0 - dble(sz2) / 6.0
                        storage = storage + dble(sz2)**2.0 + dble(sz2)
                    end if
                end if

            else
 
                call pcgetrf( szUL, szUL, gbUL(level)%ary11, 1, 1, descUL(:,1), gbIPIV(level)%ary1, info )
                if (stat /= 0) then
                    if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                        flops = flops + 2.0 * dble(szUL)**3.0 / 3.0 - dble(szUL)**2.0 / 2.0 - dble(szUL) / 6.0
                        storage = storage + dble(szUL)**2.0 + dble(szUL)
                    end if
                end if

            end if

            if (Ntmp /= lengbTT) then

                if (level <= slvl) then

                    ! compression of UR and LL
                    desc1 = descUR(:,1)
                    call gbUR_compr(gbUR(level)%ary,desc1,ctxt(level),nprow(level),npcol(level),myrow(level),mycol(level),rkmin,tol,gbUR(level)%Q1,descUR(:,1),gbUR(level)%Q2,descUR(:,2),gbUR(level)%R,descUR(:,3),gbUR(level)%rk)
                    deallocate(gbUR(level)%ary)
                    if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                        ! write(*,*) "global compr_rate: level = ", level, ", M = ", desc1(3), ", N = ", desc1(4), ", rank = ", gbUR(level)%rk
                    end if
                    if (stat /= 0) then
                        if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                            flops = flops + 4.0*dble(szUL)*dble(szLR)*dble(gbUR(level)%rk) - 2.0*dble(gbUR(level)%rk)**2.0*(dble(szUL)+dble(szLR)) & 
                                          + 4.0*dble(gbUR(level)%rk)**3.0/3.0
                            storage = storage + dble(gbUR(level)%rk) * ( dble(szUL) + dble(szLR) )
                        end if
                    end if

                    desc2 = descLL(:,1)
                    call gbUR_compr(gbLL(level)%ary,desc2,ctxt(level),nprow(level),npcol(level),myrow(level),mycol(level),rkmin,tol,gbLL(level)%Q1,descLL(:,1),gbLL(level)%Q2,descLL(:,2),gbLL(level)%R,descLL(:,3),gbLL(level)%rk)
                    deallocate(gbLL(level)%ary)
                    if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                        ! write(*,*) "global compr_rate: level = ", level, ", M = ", desc2(3), ", N = ", desc2(4), ", rank = ", gbLL(level)%rk
                    end if
                    if (stat /= 0) then
                        if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                            flops = flops + 4.0*dble(szUL)*dble(szLR)*dble(gbLL(level)%rk) - 2.0*dble(gbLL(level)%rk)**2.0*(dble(szUL)+dble(szLR)) &
                                          + 4.0*dble(gbLL(level)%rk)**3.0/3.0
                            storage = storage + dble(gbLL(level)%rk) * (dble(szUL) + dble(szLR))
                        end if
                    end if

                    ! two triangular solvers
                    call gbtrs_compr(-1,0,gbUR(level)%rk,ctxt(level),nprow(level),npcol(level),myrow(level),mycol(level),gbUL(level),descUL,gbIPIV(level),descIPIV,gbUR(level)%Q1,descUR(:,1),gbUR(level)%Q2,descUR(:,2))
                    call gbtrs_compr( 1,1,gbLL(level)%rk,ctxt(level),nprow(level),npcol(level),myrow(level),mycol(level),gbUL(level),descUL,gbIPIV(level),descIPIV,gbLL(level)%Q1,descLL(:,1),gbLL(level)%Q2,descLL(:,2))
                    if (stat /= 0) then
                        if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                            flops = flops + dble(szUL)**2.0 * ( dble(gbUR(level)%rk) + dble(gbLL(level)%rk) )
                        end if
                    end if

                    ! PBLAS 3 to update gbLR(Ntmp)%ary1
                    call gbURmm_compr(ctxt(level),nprow(level),npcol(level),myrow(level),mycol(level),gbUR(level),descUR,gbLL(level),descLL,gbLR(Ntmp)%ary1,descLR)
                    if (stat /= 0) then
                        if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                            flops = flops + dble(gbUR(level)%rk)*dble(gbLL(level)%rk)*2.0*dble(szUL) + dble(gbLL(level)%rk)*dble(szLR)*2.0*dble(gbUR(level)%rk) &
                                          + dble(szLR)*dble(szLR)*2.0*dble(gbLL(level)%rk)
                        end if
                    end if

                else

                    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  triangular solvers for gbUR and gbLL  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                    call pclapiv( 'F', 'R', 'C', szUL, szLR, gbUR(level)%ary, 1, 1, descUR, gbIPIV(level)%ary1, 1, 1, descIPIV(:,1), IWORK )
                    call pctrtrs( 'L', 'N', 'U', szUL, szLR, gbUL(level)%ary11, 1, 1, descUL(:,1), gbUR(level)%ary, 1, 1, descUR(:,1), info )
                    call pctrtrs( 'U', 'T', 'N', szUL, szLR, gbUL(level)%ary11, 1, 1, descUL(:,1), gbLL(level)%ary, 1, 1, descLL(:,1), info )
                    if (stat /= 0) then
                        if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                            flops = flops + dble(szUL)**2.0 * dble(szLR) * 2.0
                            storage = storage + dble(szUL) * dble(szLR) * 2.0
                        end if
                    end if

                    !!!!!!!!!!!!!!!!!!  PBLAS 3 updating the Schur complement which is stored at gbLR  !!!!!!!!!!!!!!!!!!!!
                    call pcgemm( 'T', 'N', szLR, szLR, szUL, -ONE, gbLL(level)%ary, 1, 1, descLL(:,1), gbUR(level)%ary, 1, 1, descUR(:,1), ONE, gbLR(Ntmp)%ary1, 1, 1, descLR )
                    if (stat /= 0) then
                        if (myrow(level) == RSRC .and. mycol(level) == CSRC) then
                            flops = flops + dble(szLR) * dble(szLR) * 2.0*dble(szUL)
                        end if
                    end if                    

                end if

            end if

        end do

        return
    end subroutine gb_fact


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  4. Parallel extend-add  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine pextadd( level, slvl, UL, descUL, UR, LL, descUR, LR, descLR, U, descU, Ntmp, kid, nbr )

        integer                     :: level, slvl, Ntmp, kid
        type(UL_struct)             :: UL
        complex(kind=4),allocatable :: UR(:,:), LL(:,:), LR(:,:), U(:,:)
        integer                     :: descUL(9,6), descUR(9), descLR(9), descU(9)
        type(NB_struct),target      :: nbr(*)

        integer                     :: sz1, sz2, szUL, length, counter, ii, jj
        integer,allocatable         :: ind(:), map(:)
        type(NB_struct),pointer     :: tailN, tailK

        ! length of the ind and map
        length = 0
        tailK => nbr(kid)%next
        do while (associated(tailK))
            length = length + 1
            tailK => tailK%next
        end do

        !! 1. UL
        if (level <= slvl) then
            sz1 = descUL(3,1)
            sz2 = descUL(3,6)
            szUL = sz1 + sz2
            call pcgeadd( 'N', sz1, sz1, ONE, U,     1,     1, descU, ONE, UL%ary11, 1, 1, descUL(:,1) )
            call pcgeadd( 'N', sz1, sz2, ONE, U,     1, sz1+1, descU, ONE, UL%ary12, 1, 1, descUL(:,2) )
            call pcgeadd( 'T', sz1, sz2, ONE, U, sz1+1,     1, descU, ONE, UL%ary21, 1, 1, descUL(:,4) )
            call pcgeadd( 'N', sz2, sz2, ONE, U, sz1+1, sz1+1, descU, ONE, UL%ary22, 1, 1, descUL(:,6) )
        else
            szUL = descUL(3,1)
            call pcgeadd( 'N', szUL, szUL, ONE, U, 1, 1, descU, ONE, UL%ary11, 1, 1, descUL(:,1) )
        end if

        if (length > 0) then

            ! ind: index of each block in U
            allocate(ind(length+1))
            ind(1) = szUL + 1
            tailK => nbr(kid)%next
            do ii = 1,length
                ind(ii+1) = ind(ii) + tailK%Otail - tailK%Ohead + 1
                tailK => tailK%next
            end do

            ! mapping to the current frontal matrix
            allocate(map(length))
            tailK => nbr(kid)%next
            do ii = 1,length
                counter = 1
                tailN => nbr(Ntmp)
                do while (associated(tailN))
                    if (tailK%Ohead >= tailN%Ohead .and. tailK%Otail <= tailN%Otail) then
                        map(ii) = tailK%Ohead - tailN%Ohead + counter
                        exit
                    end if
                    counter = counter + tailN%Otail - tailN%Ohead + 1
                    tailN => tailN%next
                end do
                tailK => tailK%next
            end do 

            !! 2. UR and LL
            do ii = 1,length
                call pcgeadd( 'N', szUL, ind(ii+1)-ind(ii), ONE, U, 1, ind(ii), descU, ONE, UR, 1, map(ii), descUR )
                call pcgeadd( 'T', szUL, ind(ii+1)-ind(ii), ONE, U, ind(ii), 1, descU, ONE, LL, 1, map(ii), descUR )
            end do

            !! 3. LR
            do jj = 1,length
                do ii = 1,length
                    call pcgeadd( 'N', ind(ii+1)-ind(ii), ind(jj+1)-ind(jj), ONE, U, ind(ii), ind(jj), descU, ONE, LR, map(ii), map(jj), descLR )
                end do
            end do

            deallocate(ind,map)

        end if

        return
    end subroutine pextadd


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  global compression  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine gbUL_compr( A, descA, ctxt, nprow, npcol, myrow, mycol, rkmin, tol, Q, descQ, R, descR, rk )

        complex(kind=4),allocatable :: A(:,:), Q(:,:), R(:,:)
        integer                     :: ctxt, nprow, npcol, myrow, mycol
        integer                     :: descA(9), descQ(9), descR(9), rkmin, rk
        real(kind=4)                :: tol

        real(kind=4)                :: R11, Rkk
        integer                     :: M, N, K, LOCR, LOCC, LOCK, Pr, Pc, ii, jj, xx, yy, pvt
        integer                     :: lwork, lrwork, info
        integer,allocatable         :: jpvt(:)
        complex(kind=4),allocatable :: tau(:), work(:)
        real(kind=4),allocatable    :: rwork(:)
        integer,external            :: numroc, indxg2p, indxg2l, indxl2g

        M = descA(3)
        N = descA(4)
        K = min(M,N)

        ! QR factorization with column pivoting
        LOCR = numroc( M, MB, myrow, RSRC, nprow )
        LOCC = numroc( N, NB, mycol, CSRC, npcol )
        LOCK = numroc( K, NB, mycol, CSRC, npcol )
        lwork = max(3,LOCR+LOCC)
        lrwork = 2*max(1,LOCC)
        allocate( jpvt(LOCC), tau(LOCK), work(lwork), rwork(lrwork) ) 
        jpvt = 0 
        call pcgeqpf( M, N, A, 1, 1, descA, jpvt, tau, work, lwork, rwork, lrwork, info )        
        deallocate(work,rwork)    

        ! broadcast R11 and Rkk
        Pr = indxg2p( 1, MB, myrow, RSRC, nprow )
        Pc = indxg2p( 1, NB, mycol, CSRC, npcol )
        if (myrow == Pr .and. mycol == Pc) then
            xx = indxg2l( 1, MB, myrow, RSRC, nprow )
            yy = indxg2l( 1, NB, mycol, CSRC, npcol )
            R11 = abs(A(xx,yy))
            call sgebs2d( ctxt, 'A', 'I', 1, 1, R11, 1 )
        else
            call sgebr2d( ctxt, 'A', 'I', 1, 1, R11, 1, Pr, Pc )
        end if

        Pr = indxg2p( rkmin, MB, myrow, RSRC, nprow )
        Pc = indxg2p( rkmin, NB, mycol, CSRC, npcol )
        if (myrow == Pr .and. mycol == Pc) then
            xx = indxg2l( rkmin, MB, myrow, RSRC, nprow )
            yy = indxg2l( rkmin, NB, mycol, CSRC, npcol )
            Rkk = abs(A(xx,yy))
            call sgebs2d( ctxt, 'A', 'I', 1, 1, Rkk, 1 )
        else
            call sgebr2d( ctxt, 'A', 'I', 1, 1, Rkk, 1, Pr, Pc )
        end if

        ! estimating rank
        if (rkmin > K) then
            rk = K
        else
            if (Rkk < tol * R11) then
                rk = rkmin
            else
                do ii = rkmin,K
                    Pr = indxg2p( ii, MB, myrow, RSRC, nprow )
                    Pc = indxg2p( ii, NB, mycol, CSRC, npcol )
                    if (myrow == Pr .and. mycol == Pc) then
                        xx = indxg2l( ii, MB, myrow, RSRC, nprow )
                        yy = indxg2l( ii, NB, mycol, CSRC, npcol )
                        if (abs(A(xx,yy)) >= tol * R11) then
                            rk = ii
                        end if
                        call igebs2d( ctxt, 'A', 'I', 1, 1, rk, 1 )
                    else
                        call igebr2d( ctxt, 'A', 'I', 1, 1, rk, 1, Pr, Pc )
                    end if
                    if (rk < ii) then
                        exit
                    end if
                end do
            end if
        end if        

        ! allocating R and copy R(1:rk,:)
        LOCR = numroc( rk, MB, myrow, RSRC, nprow )
        LOCC = numroc(  N, NB, mycol, CSRC, npcol )
        allocate(R(LOCR,LOCC))
        R = ZERO
        call descset( descR, rk, N, MB, NB, RSRC, CSRC, ctxt, max(1,LOCR) )
        do jj = 1,N
            Pc = indxg2p( jj, NB, mycol, CSRC, npcol )
            if (mycol == Pc) then
                yy = indxg2l( jj, NB, mycol, CSRC, npcol )
                pvt = jpvt(yy)
                call igebs2d( ctxt, 'R', 'I', 1, 1, pvt, 1 )
            else
                call igebr2d( ctxt, 'R', 'I', 1, 1, pvt, 1, myrow, Pc )
            end if          
            if (jj <= rk) then
                call pcgeadd( 'N', jj, 1, ONE, A, 1, jj, descA, ONE, R, 1, pvt, descR )
            else
                call pcgeadd( 'N', rk, 1, ONE, A, 1, jj, descA, ONE, R, 1, pvt, descR )
            end if
        end do        
        deallocate(jpvt)
        
        ! generate Q
        LOCR = numroc( M, MB, myrow, RSRC, nprow )
        LOCC = numroc( N, NB, mycol, CSRC, npcol )
        lwork = NB * ( NB + LOCR + LOCC )
        allocate(work(lwork))
        call pcungqr( M, rk, rk, A, 1, 1, descA, tau, work, lwork, info )
        deallocate(work,tau)

        LOCR = numroc(  M, MB, myrow, RSRC, nprow )
        LOCC = numroc( rk, NB, mycol, CSRC, npcol )
        allocate(Q(LOCR,LOCC))
        Q = ZERO
        call descset( descQ, M, rk, MB, NB, RSRC, CSRC, ctxt, max(1,LOCR) )
        call pcgeadd( 'N', M, rk, ONE, A, 1, 1, descA, ONE, Q, 1, 1, descQ )

        return
    end subroutine gbUL_compr


    subroutine gbUR_compr( A, descA, ctxt, nprow, npcol, myrow, mycol, rkmin, tol, Q1, descQ1, Q2, descQ2, R, descR, rk )

        complex(kind=4),allocatable :: A(:,:), Q1(:,:), Q2(:,:), R(:,:)
        integer                     :: ctxt, nprow, npcol, myrow, mycol
        integer                     :: descA(9), descQ1(9), descQ2(9), descR(9), rkmin, rk
        real(kind=4)                :: tol

        real(kind=4)                :: R11, Rkk
        integer                     :: M, N, K, sz1, sz2, LOCR, LOCC, LOCK, Pr, Pc, ii, jj, xx, yy, pvt
        integer                     :: lwork, lrwork, info
        integer,allocatable         :: jpvt(:)
        complex(kind=4),allocatable :: tau(:), work(:)
        real(kind=4),allocatable    :: rwork(:)
        integer,external            :: numroc, indxg2p, indxg2l, indxl2g

        M = descA(3)
        N = descA(4)
        K = min(M,N)
        sz1 = M / 2
        sz2 = M - sz1

        ! QR factorization with column pivoting
        LOCR = numroc( M, MB, myrow, RSRC, nprow )
        LOCC = numroc( N, NB, mycol, CSRC, npcol )
        LOCK = numroc( K, NB, mycol, CSRC, npcol )
        lwork = max(3,LOCR+LOCC)
        lrwork = 2*max(1,LOCC)
        allocate( jpvt(LOCC), tau(LOCK), work(lwork), rwork(lrwork) ) 
        jpvt = 0
        call pcgeqpf( M, N, A, 1, 1, descA, jpvt, tau, work, lwork, rwork, lrwork, info ) 
        deallocate(work,rwork)    

        ! broadcast R11 and Rkk
        Pr = indxg2p( 1, MB, myrow, RSRC, nprow )
        Pc = indxg2p( 1, NB, mycol, CSRC, npcol )
        if (myrow == Pr .and. mycol == Pc) then
            xx = indxg2l( 1, MB, myrow, RSRC, nprow )
            yy = indxg2l( 1, NB, mycol, CSRC, npcol )
            R11 = abs(A(xx,yy))
            call sgebs2d( ctxt, 'A', 'I', 1, 1, R11, 1 )
        else
            call sgebr2d( ctxt, 'A', 'I', 1, 1, R11, 1, Pr, Pc )
        end if

        Pr = indxg2p( rkmin, MB, myrow, RSRC, nprow )
        Pc = indxg2p( rkmin, NB, mycol, CSRC, npcol )
        if (myrow == Pr .and. mycol == Pc) then
            xx = indxg2l( rkmin, MB, myrow, RSRC, nprow )
            yy = indxg2l( rkmin, NB, mycol, CSRC, npcol )
            Rkk = abs(A(xx,yy))
            call sgebs2d( ctxt, 'A', 'I', 1, 1, Rkk, 1 )
        else
            call sgebr2d( ctxt, 'A', 'I', 1, 1, Rkk, 1, Pr, Pc )
        end if

        ! estimating rank
        if (rkmin > K) then
            rk = K
        else
            if (Rkk < tol * R11) then
                rk = rkmin
            else
                do ii = rkmin,K
                    Pr = indxg2p( ii, MB, myrow, RSRC, nprow )
                    Pc = indxg2p( ii, NB, mycol, CSRC, npcol )
                    if (myrow == Pr .and. mycol == Pc) then
                        xx = indxg2l( ii, MB, myrow, RSRC, nprow )
                        yy = indxg2l( ii, NB, mycol, CSRC, npcol )
                        if (abs(A(xx,yy)) >= tol * R11) then
                            rk = ii
                        end if
                        call igebs2d( ctxt, 'A', 'I', 1, 1, rk, 1 )
                    else
                        call igebr2d( ctxt, 'A', 'I', 1, 1, rk, 1, Pr, Pc )
                    end if
                    if (rk < ii) then
                        exit
                    end if
                end do
            end if
        end if

        ! allocating R and copy R(1:rk,:)
        LOCR = numroc( rk, MB, myrow, RSRC, nprow )
        LOCC = numroc(  N, NB, mycol, CSRC, npcol )
        allocate(R(LOCR,LOCC))
        R = ZERO
        call descset( descR, rk, N, MB, NB, RSRC, CSRC, ctxt, max(1,LOCR) )
        do jj = 1,N
            Pc = indxg2p( jj, NB, mycol, CSRC, npcol )
            if (mycol == Pc) then
                yy = indxg2l( jj, NB, mycol, CSRC, npcol )
                pvt = jpvt(yy)
                call igebs2d( ctxt, 'R', 'I', 1, 1, pvt, 1 )
            else
                call igebr2d( ctxt, 'R', 'I', 1, 1, pvt, 1, myrow, Pc )
            end if          
            if (jj <= rk) then
                call pcgeadd( 'N', jj, 1, ONE, A, 1, jj, descA, ONE, R, 1, pvt, descR )
            else
                call pcgeadd( 'N', rk, 1, ONE, A, 1, jj, descA, ONE, R, 1, pvt, descR )
            end if
        end do        
        deallocate(jpvt)
        
        ! generate Q1 and Q2
        LOCR = numroc( M, MB, myrow, RSRC, nprow )
        LOCC = numroc( N, NB, mycol, CSRC, npcol )
        lwork = NB * ( NB + LOCR + LOCC )
        allocate(work(lwork))
        call pcungqr( M, rk, rk, A, 1, 1, descA, tau, work, lwork, info )
        deallocate(work,tau)

        LOCR = numroc( sz1, MB, myrow, RSRC, nprow )
        LOCC = numroc(  rk, NB, mycol, CSRC, npcol )
        allocate(Q1(LOCR,LOCC))
        Q1 = ZERO
        call descset( descQ1, sz1, rk, MB, NB, RSRC, CSRC, ctxt, max(1,LOCR) )

        LOCR = numroc( sz2, MB, myrow, RSRC, nprow )
        LOCC = numroc(  rk, NB, mycol, CSRC, npcol )
        allocate(Q2(LOCR,LOCC))
        Q2 = ZERO
        call descset( descQ2, sz2, rk, MB, NB, RSRC, CSRC, ctxt, max(1,LOCR) )

        call pcgeadd( 'N', sz1, rk, ONE, A,     1, 1, descA, ONE, Q1, 1, 1, descQ1 )
        call pcgeadd( 'N', sz2, rk, ONE, A, sz1+1, 1, descA, ONE, Q2, 1, 1, descQ2 )

        return
    end subroutine gbUR_compr


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  global UR and LL multiplication  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine gbULmm_compr( ctxt, nprow, npcol, myrow, mycol, Q1, descQ1, R1, descR1, Q2, descQ2, R2, descR2, A, descA )

        integer                     :: ctxt, nprow, npcol, myrow, mycol
        complex(kind=4),allocatable :: Q1(:,:), R1(:,:), Q2(:,:), R2(:,:), A(:,:)
        integer                     :: descQ1(9), descR1(9), descQ2(9), descR2(9), descA(9)

        integer                     :: sz1, sz2, rk1, rk2, LOCR, LOCC
        integer                     :: descTMP1(9), descTMP2(9)
        complex(kind=4),allocatable :: tmp1(:,:), tmp2(:,:)
        integer,external            :: numroc
        

        sz1 = descQ1(3)
        sz2 =  descA(3)
        rk1 = descR1(3)
        rk2 = descR2(3)

        ! tmp1
        LOCR = numroc( rk2, MB, myrow, RSRC, nprow )
        LOCC = numroc( rk1, NB, mycol, CSRC, npcol )
        allocate(tmp1(LOCR,LOCC))
        tmp1 = ZERO
        call descset( descTMP1, rk2, rk1, MB, NB, RSRC, CSRC, ctxt, max(1,LOCR) )

        ! tmp2
        LOCR = numroc( rk2, MB, myrow, RSRC, nprow )
        LOCC = numroc( sz2, NB, mycol, CSRC, npcol )
        allocate(tmp2(LOCR,LOCC))
        tmp2 = ZERO
        call descset( descTMP2, rk2, sz2, MB, NB, RSRC, CSRC, ctxt, max(1,LOCR) ) 

        call pcgemm( 'T', 'N', rk2, rk1, sz1, ONE, Q2, 1, 1, descQ2, Q1, 1, 1, descQ1, ONE, tmp1, 1, 1, descTMP1 )
        call pcgemm( 'N', 'N', rk2, sz2, rk1, ONE, tmp1, 1, 1, descTMP1, R1, 1, 1, descR1, ONE, tmp2, 1, 1, descTMP2 )
        deallocate(tmp1)
        call pcgemm( 'T', 'N', sz2, sz2, rk2, -ONE, R2, 1, 1, descR2, tmp2, 1, 1, descTMP2, ONE, A, 1, 1, descA )
        deallocate(tmp2)

        return
    end subroutine gbULmm_compr


    subroutine gbURmm_compr( ctxt, nprow, npcol, myrow, mycol, UR, descUR, LL, descLL, A, descA )

        integer                     :: ctxt, nprow, npcol, myrow, mycol
        type(UR_struct)             :: UR, LL
        complex(kind=4),allocatable :: A(:,:)
        integer                     :: descUR(9,3), descLL(9,3), descA(9)

        integer                     :: sz1, sz2, rkUR, rkLL, szLR, LOCR, LOCC
        integer                     :: descTMP1(9), descTMP2(9)
        complex(kind=4),allocatable :: tmp1(:,:), tmp2(:,:)
        integer,external            :: numroc
        

        sz1  = descUR(3,1)
        sz2  = descUR(3,2)
        rkUR = UR%rk
        rkLL = LL%rk 
        szLR = descA(3)

        ! tmp1
        LOCR = numroc( rkLL, MB, myrow, RSRC, nprow )
        LOCC = numroc( rkUR, NB, mycol, CSRC, npcol )
        allocate(tmp1(LOCR,LOCC))
        tmp1 = ZERO
        call descset( descTMP1, rkLL, rkUR, MB, NB, RSRC, CSRC, ctxt, max(1,LOCR) )

        ! tmp2
        LOCR = numroc( rkLL, MB, myrow, RSRC, nprow )
        LOCC = numroc( szLR, NB, mycol, CSRC, npcol )
        allocate(tmp2(LOCR,LOCC))
        tmp2 = ZERO
        call descset( descTMP2, rkLL, szLR, MB, NB, RSRC, CSRC, ctxt, max(1,LOCR) ) 

        call pcgemm( 'T', 'N', rkLL, rkUR, sz1,  ONE, LL%Q1, 1, 1, descLL(:,1), UR%Q1, 1, 1, descUR(:,1), ONE, tmp1, 1, 1, descTMP1 )
        call pcgemm( 'T', 'N', rkLL, rkUR, sz2,  ONE, LL%Q2, 1, 1, descLL(:,2), UR%Q2, 1, 1, descUR(:,2), ONE, tmp1, 1, 1, descTMP1 )
        call pcgemm( 'N', 'N', rkLL, szLR, rkUR,  ONE, tmp1, 1, 1, descTMP1, UR%R, 1, 1, descUR(:,3), ONE, tmp2, 1, 1, descTMP2 )
        deallocate(tmp1)
        call pcgemm( 'T', 'N', szLR, szLR, rkLL, -ONE, LL%R, 1, 1, descLL(:,3), tmp2, 1, 1, descTMP2, ONE, A, 1, 1, descA )
        deallocate(tmp2)

        return
    end subroutine gbURmm_compr


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  global solution to one RHS  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine gbtrs_compr( UPLO, TRANS, NRHS, ctxt, nprow, npcol, myrow, mycol, UL, descUL, IPIV, descIPIV, B1, descB1, B2, descB2 )

        integer                     :: UPLO, TRANS, NRHS, ctxt, nprow, npcol, myrow, mycol
        type(UL_struct)             :: UL
        type(IPIV_struct)           :: IPIV
        complex(kind=4),allocatable :: B1(:,:), B2(:,:)
        integer                     :: descUL(9,6), descIPIV(9,2), descB1(9), descB2(9)

        integer                     :: sz1, sz2, rk1, rk2, LOCR, LOCC, info
        complex(kind=4),allocatable :: tmp(:,:)
        integer                     :: descTMP(9), IWORK(1)
        integer,external            :: numroc

        sz1 = descUL(3,1)
        sz2 = descUL(3,6)
        rk1 = descUL(3,3)
        rk2 = descUL(3,5) 

        if (UPLO < 0) then        ! lower triangular system

            call pclapiv( 'F', 'R', 'C', sz1, NRHS, B1, 1, 1, descB1, IPIV%ary1, 1, 1, descIPIV(:,1), IWORK )
            call pctrtrs( 'L', 'N', 'U', sz1, NRHS, UL%ary11, 1, 1, descUL(:,1), B1, 1, 1, descB1, info )

            LOCR = numroc(  rk2, MB, myrow, RSRC, nprow )
            LOCC = numroc( NRHS, NB, mycol, CSRC, npcol )
            allocate(tmp(LOCR,LOCC))
            tmp = ZERO
            call descset( descTMP, rk2, NRHS, MB, NB, RSRC, CSRC, ctxt, max(1,LOCR) )
            call pcgemm( 'T', 'N', rk2, NRHS, sz1,  ONE, UL%Q2, 1, 1, descUL(:,4), B1, 1, 1, descB1, ONE, tmp, 1, 1, descTMP )
            call pcgemm( 'T', 'N', sz2, NRHS, rk2, -ONE, UL%R2, 1, 1, descUL(:,5), tmp, 1, 1, descTMP, ONE, B2, 1, 1, descB2 ) 
            deallocate(tmp)

            call pclapiv( 'F', 'R', 'C', sz2, NRHS, B2, 1, 1, descB2, IPIV%ary2, 1, 1, descIPIV(:,2), IWORK )
            call pctrtrs( 'L', 'N', 'U', sz2, NRHS, UL%ary22, 1, 1, descUL(:,6), B2, 1, 1, descB2, info )

        elseif (UPLO > 0) then     ! upper triangular system

            if (TRANS == 0) then    ! NO transpose

                call pctrtrs( 'U', 'N', 'N', sz2, NRHS, UL%ary22, 1, 1, descUL(:,6), B2, 1, 1, descB2, info )

                LOCR = numroc(  rk1, MB, myrow, RSRC, nprow )
                LOCC = numroc( NRHS, NB, mycol, CSRC, npcol )
                allocate(tmp(LOCR,LOCC))
                tmp = ZERO
                call descset( descTMP, rk1, NRHS, MB, NB, RSRC, CSRC, ctxt, max(1,LOCR) )
                call pcgemm( 'N', 'N', rk1, NRHS, sz2,  ONE, UL%R1, 1, 1, descUL(:,3), B2, 1, 1, descB2, ONE, tmp, 1, 1, descTMP )
                call pcgemm( 'N', 'N', sz1, NRHS, rk1, -ONE, UL%Q1, 1, 1, descUL(:,2), tmp, 1, 1, descTMP, ONE, B1, 1, 1, descB1 )
                deallocate(tmp)

                call pctrtrs( 'U', 'N', 'N', sz1, NRHS, UL%ary11, 1, 1, descUL(:,1), B1, 1, 1, descB1, info )

            else                    ! YES transpose

                call pctrtrs( 'U', 'T', 'N', sz1, NRHS, UL%ary11, 1, 1, descUL(:,1), B1, 1, 1, descB1, info )

                LOCR = numroc(  rk1, MB, myrow, RSRC, nprow )
                LOCC = numroc( NRHS, NB, mycol, CSRC, npcol )
                allocate(tmp(LOCR,LOCC))
                tmp = ZERO
                call descset( descTMP, rk1, NRHS, MB, NB, RSRC, CSRC, ctxt, max(1,LOCR) )
                call pcgemm( 'T', 'N', rk1, NRHS, sz1,  ONE, UL%Q1, 1, 1, descUL(:,2), B1, 1, 1, descB1, ONE, tmp, 1, 1, descTMP )
                call pcgemm( 'T', 'N', sz2, NRHS, rk1, -ONE, UL%R1, 1, 1, descUL(:,3), tmp, 1, 1, descTMP, ONE, B2, 1, 1, descB2 ) 
                deallocate(tmp)

                call pctrtrs( 'U', 'T', 'N', sz2, NRHS, UL%ary22, 1, 1, descUL(:,6), B2, 1, 1, descB2, info )

            end if

        end if

        return
    end subroutine gbtrs_compr


end module fact
