module partition

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!  Nested dissection preconditioning, applicable to both 2D and 3D  !!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
implicit none

    !! elimination tree: initializing to be zero
    type :: TT_struct
        integer :: LL      = 0
        integer :: parent  = 0
        integer :: kids(2) = 0 
    end type TT_struct

    !! TL structure
    type :: TL_struct
        integer,allocatable :: ind(:), ary(:)
    end type TL_struct

    !! NB pointer
    type :: NB_struct
        integer                 :: Node, Ohead, Otail   
        type(NB_struct),pointer :: next
    end type NB_struct

    !! CSR: Compressed Sparse Row storage of A
    type :: CSR
        integer,allocatable         :: ind(:)
        integer,allocatable         :: col(:)
        complex(kind=4),allocatable :: val(:)
    end type CSR 

    !! global variables
    integer                     :: lengbTT, lenlocTT, flag, locHX, locTX, locHY, locTY, locHZ, locTZ
    type(TT_struct),allocatable :: gbTT(:), locTT(:)
    integer,allocatable         :: gbTlabel(:), locTlabel(:)
    type(NB_struct),allocatable :: gbNB(:), locNB(:)
    type(CSR)                   :: A
    complex(kind=4),parameter   :: pi = (3.1415926E+0,0.0E+0)
    complex(kind=4),parameter   :: ONE = (1.0E+0,0.0E+0), ZERO = (0.0E+0,0.0E+0), RT = (0.0E+0,1.0E+0)


contains


    !!!!!!!!!!!!!!!!!!!!!!  1. elimination tree generation  !!!!!!!!!!!!!!!!!!!
    subroutine TT_generate( lenTT, TT )

        integer                     :: lenTT
        type(TT_struct),allocatable :: TT(:)

        integer                     :: L1, L2, level, ind, counter, Node, Ntmp
        integer,allocatable         :: nop(:,:)
        type :: tmp_struct
            integer                  :: val
            type(tmp_struct),pointer :: prev, next
        end type tmp_struct
        type(tmp_struct),pointer     :: head, tail

        ! level bounded by L1 and L2: DO NOT use floor and ceiling, not numerically stable
        L1 = int(log(real(lenTT+1))/log(2.0))
        L2 = L1+1
        if (lenTT+1 == 2**L1) then
            L2 = L1
        elseif (lenTT+1 == 2**L2) then
            L1 = L2
        end if

        ! allocating tree
        allocate(TT(lenTT))

        ! building up TT
        allocate(nop(L2,3))
        nop = 0
        do Ntmp = 1,lenTT

            ind = 0
            do level = 1,L2
                if (nop(level,1) == 2) then
                    ind = level    ! search for the level which has two no-parent nodes
                    exit
                end if    
            end do

            if (ind == 0) then    ! starting from the bottom level
                level = L2
                nop(level,1) = nop(level,1) + 1
                if ( nop(level,2)==0 ) then
                    nop(level,2) = Ntmp
                else
                    nop(level,3) = Ntmp
                end if
            else
                level = ind - 1    ! go to the upper level to form the tree
                nop(level,1) = nop(level,1) + 1
                if ( nop(level,2) == 0 ) then
                    nop(level,2) = Ntmp
                else
                    nop(level,3) = Ntmp
                end if
                TT(Ntmp)%kids = nop(ind,2:3)
                TT( nop(ind,2) )%parent = Ntmp
                TT( nop(ind,3) )%parent = Ntmp
                nop(ind,:) = 0    ! clear step: start over
            end if

        end do
        deallocate(nop)

        ! modify the tree if L1 /= L2
        if (L1 /= L2) then

            ! initial count: how many nodes do they have parent 0
            head => null()
            tail => null()
            counter = 0
            do Ntmp = 1,lenTT
                if (TT(Ntmp)%parent == 0) then
                    counter = counter + 1
                    if (counter == 1) then
                        allocate(head)
                        head%val  = Ntmp
                        head%prev => null()
                        head%next => null()
                        tail => head
                    else
                        allocate(tail%next)
                        tail%next%val = Ntmp
                        tail%next%prev => tail
                        tail%next%next => null()
                        tail => tail%next
                    end if
                end if
            end do   ! up to now, counter is the length of tmp

            ! modification
            do while (counter > 1)
                Node = tail%val
                if (TT(Node)%kids(1) /= 0) then
                    tail%val = TT(Node)%kids(1)
                    counter = counter + 1
                    allocate(tail%next)
                    tail%next%val = TT(Node)%kids(2)
                    tail%next%prev => tail
                    tail%next%next => null()
                    tail => tail%next
                else
                    tail => tail%prev
                    counter = counter - 1
                    deallocate(tail%next)
                    tail%next => null()
                end if
                TT(Node)%kids(1) = head%val
                TT(Node)%kids(2) = tail%val
                TT( head%val )%parent = Node
                TT( tail%val )%parent = Node
                head => head%next
                counter = counter - 1
                deallocate(head%prev)
                head%prev => null()
            end do
            deallocate(head)
            head => null()
            tail => null()
 
        end if

        ! set up level information
        TT(lenTT)%LL = 1
        do Ntmp = lenTT-1,1,-1
            TT(Ntmp)%LL = TT( TT(Ntmp)%parent )%LL + 1
        end do

        return
    end subroutine TT_generate


    !!!!!!!!!!!!!!!!!!!!!!!!!!!  2. TL_generate  !!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine TL_generate( lenTT, TT, TL )

        integer         :: lenTT
        type(TT_struct) :: TT(lenTT)
        type(TL_struct) :: TL

        integer             :: Lvl, level, Ntmp
        integer,allocatable :: counter(:)

        Lvl = TT(1)%LL
        allocate( counter(Lvl), TL%ind(Lvl+1), TL%ary(lenTT) )
        counter = 0
        TL%ind  = 0
        TL%ary  = 0

        ! compute how many nodes are in the same level
        do Ntmp = 1,lenTT
            level = TT(Ntmp)%LL
            counter(level) = counter(level) + 1
        end do
        TL%ind(1) = 1
        do level = 1,Lvl
            TL%ind(level+1) = TL%ind(level) + counter(level)
        end do

        ! fill the ary
        counter = TL%ind(1:Lvl) - 1
        do Ntmp = 1,lenTT
            level = TT(Ntmp)%LL
            counter(level) = counter(level) + 1
            TL%ary( counter(level) ) = Ntmp
        end do
        deallocate(counter)
        
        return
    end subroutine TL_generate


end module partition
